from __future__ import annotations

import typing as t
from typing_extensions import Self


class IdentityMappingProtocol(t.Protocol):
    DATA_TYPE: str

    @classmethod
    def from_mapping(
        cls,
        mapping_document: t.Mapping,
        connector: str | None = None,
        storage_gateway: str | None = None,
    ) -> Self:
        """
        Instantiate an identity mapper from the provided mapping
        configuration document.
        """

    @classmethod
    def map_identity(self, identity_data: t.Mapping[str, str]) -> str | None:
        """
        Map an identity resource to a connector-specific username.
        """

    def map_identities(
        self, identity_data: t.Iterable[t.Mapping[str, str]]
    ) -> list[dict[str, list[str]]]:
        """
        Map a list of identity resources to connector-specific usernames.
        """



import typing as t


def create_patch_schema(schema_class, **kwargs):
    """
    Instantiate a schema class for use with PATCH operations. Adds the
    following behaviors:

    -   Sets partial=True at the schema so that missing fields are allowed
    -   Sets the allow_none property of each field to True
    """
    if "partial" in kwargs:
        del kwargs["partial"]

    new_schema = schema_class(partial=True, **kwargs)
    for f in new_schema.fields:
        new_schema.fields[f].allow_none = True

    return new_schema


class IdentityMappingConfiguration(t.TypedDict):
    DATA_TYPE: str



from __future__ import annotations

import sys
import typing as t

import marshmallow as ma
import marshmallow.validate
import marshmallow_oneofschema  # type: ignore

from . import create_patch_schema
from .schema_version_checker import verify_schema_version, version_safe_type_schemas


class DataTypeField(ma.fields.String):
    def __init__(self, value, *args, **kwargs) -> None:
        self._DATA_TYPE = value
        super().__init__(
            *args,
            dump_default=value,
            required=True,
            validate=ma.validate.Equal(value),
            metadata={"description": "Type of this document"},
            **kwargs,
        )


class OneOfSchema(marshmallow_oneofschema.OneOfSchema):
    _DATA_TYPE: t.ClassVar[str]
    private_schema: t.ClassVar[OneOfSchema]
    public_schema: t.ClassVar[OneOfSchema]
    patch_schema: t.ClassVar[OneOfSchema]

    metadata: t.ClassVar[t.Dict[str, t.Any]]

    private: t.ClassVar[t.List[str]] = []


class Schema(ma.Schema):
    _DATA_TYPE: t.ClassVar[str]
    DATA_TYPE: t.ClassVar[DataTypeField]

    metadata: t.ClassVar[t.Dict[str, t.Any]]

    @ma.validates("DATA_TYPE")
    def validate_data_type(self, data, **kwargs):
        verify_schema_version(data, self._DATA_TYPE)

    private_schema: t.ClassVar[Schema]
    public_schema: t.ClassVar[Schema]
    patch_schema: t.ClassVar[Schema]

    private: t.ClassVar[t.List[str]] = []


if sys.version_info < (3, 9):
    _type_schema = version_safe_type_schemas()
else:
    # After pyupgrade and there is no < (3, 9) branch, _type_schema can
    # be reintegrated below.
    _type_schema = version_safe_type_schemas[str, ma.Schema]()


class DataTypeSchema(marshmallow_oneofschema.OneOfSchema):
    type_field = "DATA_TYPE"
    type_schemas: t.ClassVar = _type_schema
    type_field_remove = False
    __oneof_schemas: t.ClassVar = {}

    def __getattr__(self, name):
        if name in self.__class__.__oneof_schemas:
            return self.__class__.__oneof_schemas[name]
        else:
            raise AttributeError(name)

    @property
    def schemas(self) -> t.List[Schema]:
        return list(self.__class__.__oneof_schemas.values())

    @classmethod
    def register_schema(
        cls,
        schema: Schema,
        register_data_type: bool = True,
        schema_base_name: t.Optional[str] = None,
        private: bool = False,
    ) -> None:
        DATA_TYPE = schema._DATA_TYPE

        if register_data_type:
            cls.type_schemas[DATA_TYPE] = schema

        base_data_type, version = DATA_TYPE.split("#", 1)

        if schema_base_name:
            base_name = schema_base_name
        else:
            base_name = base_data_type.title().replace("_", "")
        major, minor, patch = (int(i) for i in version.split("."))

        if major == 1:
            base_name_key = base_name
        else:
            base_name_key = f"{base_name}_{major}"

        if base_name_key not in cls.__oneof_schemas:

            class data_type_oneof(OneOfSchema):
                __doc__ = schema.__doc__
                _DATA_TYPE: t.ClassVar[str] = DATA_TYPE
                type_field = "DATA_TYPE"
                type_schemas = _type_schema
                type_field_remove = False
                metadata: t.ClassVar[t.Dict[str, t.Any]] = {}

                def get_obj_type(self, obj):
                    return self._DATA_TYPE

            data_type_oneof.__name__ = base_name_key
            instance = data_type_oneof(load_only=schema.private)
            cls.__oneof_schemas[base_name_key] = instance

            class data_type_oneof_private(OneOfSchema):
                __doc__ = schema.__doc__
                _DATA_TYPE: t.ClassVar[str] = DATA_TYPE
                type_field = "DATA_TYPE"
                type_schemas = _type_schema
                type_field_remove = False
                metadata: t.ClassVar[t.Dict[str, t.Any]] = {}

                def get_obj_type(self, obj):
                    return self._DATA_TYPE

            data_type_oneof_private.__name__ = f"{base_name_key}Private"

            class data_type_oneof_patch(OneOfSchema):
                __doc__ = schema.__doc__
                _DATA_TYPE: t.ClassVar[str] = DATA_TYPE
                type_field = "DATA_TYPE"
                type_schemas = _type_schema
                type_field_remove = False
                metadata: t.ClassVar[t.Dict[str, t.Any]] = {}

                def get_obj_type(self, obj):
                    return self._DATA_TYPE

            data_type_oneof_patch.__name__ = f"{base_name_key}Patch"

            if private:
                data_type_oneof.metadata["x-private"] = True
                data_type_oneof_private.metadata["x-private"] = True
                data_type_oneof_patch.metadata["x-private"] = True

            data_type_oneof.public_schema = instance
            data_type_oneof.private_schema = data_type_oneof_private()
            data_type_oneof.patch_schema = data_type_oneof_patch(partial=True)
            cls.__oneof_schemas[base_name_key] = data_type_oneof()

        oo = cls.__oneof_schemas[base_name_key]
        oo.type_schemas[DATA_TYPE] = schema
        for private_prop in schema.private:
            oo.load_only.add(private_prop)
        oo.private_schema.__class__.type_schemas[DATA_TYPE] = schema.private_schema
        oo.patch_schema.__class__.type_schemas[DATA_TYPE] = schema.patch_schema

        old_major, old_minor, old_patch = (
            int(i) for i in oo._DATA_TYPE.split("#", 1)[1].split(".")
        )
        if (
            major > old_major
            or (major == old_major and minor > old_minor)
            or (major == old_major and minor == old_minor and patch > old_patch)
        ):
            oo._DATA_TYPE = DATA_TYPE
            oo.private_schema._DATA_TYPE = DATA_TYPE
            oo.public_schema._DATA_TYPE = DATA_TYPE
            oo.public_schema.private = schema.private
            oo.patch_schema._DATA_TYPE = DATA_TYPE

            oo.__doc__ = schema.__doc__
            oo.private_schema.__doc__ = schema.__doc__
            oo.patch_schema.__doc__ = schema.__doc__

    @classmethod
    def schema_for(cls, DATA_TYPE) -> Schema:
        return cls.type_schemas[DATA_TYPE]

    @classmethod
    def schema_for_base_name(cls, base_name: str) -> OneOfSchema:
        return cls.__oneof_schemas[base_name]


data_type_schema = DataTypeSchema()


T = t.TypeVar("T", bound=t.Type[Schema])


def data_type(
    discriminator: str,
    register_data_type: bool = True,
    private: bool = False,
    schema_base_name: t.Optional[str] = None,
) -> t.Callable[[T], T]:
    def decorator(cls: T) -> T:
        cls._DATA_TYPE = discriminator
        cls.DATA_TYPE = DataTypeField(discriminator)
        cls._declared_fields["DATA_TYPE"] = cls.DATA_TYPE
        cls.private_schema = cls()
        cls.public_schema = cls(load_only=cls.private)
        cls.patch_schema = create_patch_schema(cls)
        if not hasattr(cls, "metadata"):
            cls.metadata = {}

        if private:
            cls.metadata["x-private"] = True

        inst = cls.public_schema
        DataTypeSchema.register_schema(
            inst,
            register_data_type=register_data_type,
            schema_base_name=schema_base_name,
            private=private,
        )
        return cls

    return decorator



from __future__ import annotations

import sys
import typing as t
from collections import UserDict

import marshmallow as ma

_message = (
    'The endpoint does not recognize doc {} "{}". Either your '
    "client is misconfigured or the service requires an update to "
    "support it."
)


def _doc_version_error(data_type: str) -> str:
    return _message.format("version", data_type)


def _doc_type_error(data_type: str) -> str:
    return _message.format("type", data_type)


def _do_prefixes_match(s1: str, s2: str, delim: str = "#") -> bool:
    fields1 = s1.split(delim)
    fields2 = s2.split(delim)

    # For our use case, an empty prefix (doc type) is not legal and so we
    # can safely say they do not match
    if len(fields1) == 0 or len(fields1[0]) == 0:
        return False
    if len(fields2) == 0 or len(fields2[0]) == 0:
        return False
    return fields1[0] == fields2[0]


def verify_schema_version(data_type_in: str, data_type_expected: str) -> None:
    # Simple case, exact match
    if data_type_in == data_type_expected:
        return
    # doc type is not recognized
    if not _do_prefixes_match(data_type_in, data_type_expected):
        raise ma.ValidationError(_doc_type_error(data_type_in))
    raise ma.ValidationError(_doc_version_error(data_type_in))


_K = t.TypeVar("_K", bound=str)
_V = t.TypeVar("_V", bound=ma.Schema)

if sys.version_info < (3, 9):
    if t.TYPE_CHECKING:
        TypedUserDict = UserDict[_K, _V]
    else:
        TypedUserDict = UserDict

    class version_safe_type_schemas(TypedUserDict):
        def __getitem__(self, key: _K) -> _V:
            # Simple case, key exists
            if key in self.data:
                return self.data[key]
            verify_schema_version(key, list(self.data.keys())[0])
            raise KeyError(key)

else:

    class version_safe_type_schemas(UserDict[_K, _V]):
        """Doc type and version checking for marshmallow OneOfSchema

        marshamallow OneOfSchema uses the dict field type_schemas
        to demarshal polymorphic schema. This class behaves the same
        as type_schemas with the added benefit of errors messages more
        specific to the type of error (wrong doc type or version).

        Note: We do assume that the doc_type's in type_schema are all
        identical, so we just the first key as the prefix to check.
        """

        def __getitem__(self, key: _K) -> _V:
            # Simple case, key exists
            if key in self.data:
                return self.data[key]
            verify_schema_version(key, list(self.data.keys())[0])
            raise KeyError(key)



from __future__ import annotations

import typing as t
from abc import ABC, abstractmethod
from typing_extensions import Self


class IdentityMapping(ABC):
    DATA_TYPE: str

    @classmethod
    def from_mapping_document(
        cls,
        mapping_document: t.Mapping,
        connector: str | None = None,
        storage_gateway: str | None = None,
    ) -> Self:
        """
        Instantiate an identity mapper from the provided mapping
        configuration document.
        """
        if mapping_document.get("DATA_TYPE") != cls.DATA_TYPE:
            raise ValueError("Unsupported DATA_TYPE")

        return cls._from_mapping_document(mapping_document, connector, storage_gateway)

    @classmethod
    @abstractmethod
    def _from_mapping_document(
        cls,
        mapping_document: t.Any,
        connector: str | None,
        storage_gateway: str | None,
    ):
        raise NotImplementedError



from __future__ import annotations

import logging
import shlex
import typing as t

from subprocess import PIPE, Popen

import marshmallow as ma
import marshmallow.validate
from typing_extensions import Self

from .abstract import IdentityMapping
from ..errors import IdentityMappingError, InvalidMappingError
from ..schema import IdentityMappingConfiguration


class MappingInputSchema(ma.Schema):
    _DATA_TYPE = "identity_mapping_input#1.0.0"
    DATA_TYPE = ma.fields.Str(
        dump_default=_DATA_TYPE, validate=[ma.validate.Equal(_DATA_TYPE)]
    )

    identities = ma.fields.List(ma.fields.Raw)


class MappingResultSchema(ma.Schema):
    id = ma.fields.UUID()
    output = ma.fields.Str()


class MappingOutputSchema(ma.Schema):
    _DATA_TYPE = "identity_mapping_output#1.0.0"
    DATA_TYPE = ma.fields.Str(
        dump_default=_DATA_TYPE, validate=[ma.validate.Equal(_DATA_TYPE)]
    )
    result = ma.fields.List(ma.fields.Nested(MappingResultSchema))


class ExternalMappingConfigurationDocument(IdentityMappingConfiguration):
    command: list[str | int | float]


identity_mapping_input_schema = MappingInputSchema()
identity_mapping_output_schema = MappingOutputSchema()


class ExternalIdentityMapping(IdentityMapping):
    """
    The ExternalIdentityMapping class implements an extension point for
    administrators to implement their own identity mapping algorithm as an
    external program.

    The GCS Manager will run the configured program and arguments in a sub
    process. In addition to any arguments specified in the configuration
    document, the following command-line options will be added:

    The `-c` option along with an argument which is the UUID of a
    GCS connector.

    The `-s` option along with its argument which is the UUID of the
    storage gateway mapping the identity.

    The `-a` command-line option indicates that the GCS Manager wants to
    receive output containing all mappings for the given identity set. If
    present, the program will receive an input containing at least one identity
    resource as input, and may return as many mappings as can be computed from
    the identities. If not present, the program will receive exactly one object
    in the identities orray and may only return a single mapping for that
    identity. If it returns multiple mappings, only the first is used by
    the GCS Manager.

    The GCS Manager communicates between the external program using
    JSON-formatted input and output data passed via stdin and stdout.

    Input consists of a JSON object containing two properties: `DATA_TYPE`
    which indicates the version of the input document, currently
    `identity_mapping_input#1.0.0` and `identities`, which contains a list of
    identity resource obtained from Globus Auth.

    For example:

    .. code-block:: json

        {
            "DATA_TYPE": "identity_mapping_input#1.0.0",
            "identities": [
                {
                    "id": "9cf5a1d2-41c4-4925-b0bb-f5e0a16c98ab",
                    "username": "user@example.com",
                    "identity_provider": "1f8ba017-6224-43ad-a42f-03059574ba38"
                },
                {
                    "id": "67c31064-1e5f-4d78-b687-5fbea3d3e05a",
                    "username": "user2@example.org",
                    "identity_provider": "05f6f886-9277-42b9-836e-81559f337ac3"
                },
                {
                    "id": "4787931e-b999-412e-97f9-a7a24c79ad71",
                    "username": "user3@example.info",
                    "identity_provider": "05f6f886-9277-42b9-836e-81559f337ac3"
                }
            ]
        }

    Output is an object containing two properties, `DATA_TYPE` which
    indicates the version of the output document, currently
    `identity_mapping_output#1.0.0`, and `result`, which contains a list of
    objects with the properties `id` and `output` The `id` is the value of one
    of the `id` properties of one of the input identities, and the `output` is
    a connector-specific user identifier that indicates the result of the
    mapping.

    For example:

    .. code-block:: json

        {
            "DATA_TYPE": "identity_mapping_output#1.0.0",
            "result": [
                {
                    "id": "67c31064-1e5f-4d78-b687-5fbea3d3e05a",
                    "output": "user3"
                },
            ]
        }
    """

    DATA_TYPE = "external_identity_mapping#1.0.0"
    mapping_timeout_s = 5  # (seconds) time limit for external program to respond

    def __init__(
        self,
        cmd: str,
        *cmd_args,
        storage_gateway: str,
        connector: str,
    ) -> None:
        """
        :param `*args`:
            List of strings which are the path to the external command and the
            command line arguments to pass to the module. Additional parameters
            `-s`, `-c`, and `-a` will be automatically added to this
            command-line as described in the class documentation.

        :param storage_gateway:
            String containing the UUID of the storage gateway to perform the
            mapping for.
        :param connector: String containing the UUID of the connector
            to perform the mapping for.
        """
        assert connector is not None
        assert storage_gateway is not None

        self._log: logging.Logger | None = None

        self.command = [str(cmd)]
        self.command.extend(map(str, cmd_args))
        self.connector = connector
        self.storage_gateway = storage_gateway

    def __repr__(self) -> str:
        args = "', '".join(self.command)
        sg, conn = self.storage_gateway, self.connector
        sg = "storage_gateway=None" if sg is None else f"storage_gateway='{sg}'"
        conn = "connector=None" if conn is None else f"connector='{conn}'"
        return f"{type(self).__name__}('{args}', {sg}, {conn})"

    @classmethod
    def _from_mapping_document(
        cls,
        mapping_document: ExternalMappingConfigurationDocument,
        connector: str | None = None,
        storage_gateway: str | None = None,
    ) -> Self:
        # The protocol allows for connector and storage_gateway to be None,
        # but this class demands it; so `assert` instead of changing the sig
        assert connector is not None, "Connector is required"
        assert storage_gateway is not None, "Connector is required"
        return cls(
            str(mapping_document["command"][0]),
            *mapping_document["command"][1:],
            connector=connector,  # noqa: arg-type
            storage_gateway=storage_gateway,
        )

    def map_identity(self, identity_data: t.Mapping[str, str]) -> str | None:
        """
        Map a single identity object to a username for this particular
        connector and storage gateway.

        :param identity_data:
            A dict containing the identity assertions from Globus Auth.

        :raises errors.InternalServerError:
            Error executing the external mapping program, or the output
            from the program is not in an understood format.
        """
        cmd = self.command + ["-c", self.connector, "-s", self.storage_gateway]
        cmd_str = shlex.join(cmd)
        p = Popen(cmd, universal_newlines=True, stdout=PIPE, stderr=PIPE, stdin=PIPE)

        input_data = identity_mapping_input_schema.dumps(
            {"identities": [identity_data]}
        )

        stdout, stderr = p.communicate(input_data, timeout=self.mapping_timeout_s)
        self.log.debug(
            # fmt: off
            "\n  command: %s"
            "\n    stdin: %s"
            "\n   stdout: %s"
            "\n   stderr: %s",
            # fmt: on
            cmd_str,
            input_data,
            stdout,
            stderr,
        )

        if p.returncode != 0:
            msg = (
                "Error mapping identity\n"
                f"    command: {cmd_str}\n"
                f"  exit code: {p.returncode}\n"
                f"      stdin: {input_data}\n"
                f"     stdout: {stdout}\n"
                f"     stderr: {stderr}"
            )
            raise IdentityMappingError(msg)

        if stdout:
            try:
                output = identity_mapping_output_schema.loads(stdout)
            except Exception as e:
                msg = (
                    "The mapping application yielded an invalid result:\n"
                    f"  exception: ({type(e).__name__}) {e}\n"
                    f"    command: {cmd_str}\n"
                    f"      stdin: {input_data}\n"
                    f"     stdout: {stdout}\n"
                    f"     stderr: {stderr}"
                )
                raise IdentityMappingError(msg) from e

            if output["result"]:
                first_result = output["result"][0]
                if str(first_result["id"]) != identity_data["id"]:
                    msg = (
                        "Expected first result to match identity data (id field);"
                        " assuming response is invalid.\n"
                        f"  command: {cmd_str}\n"
                        f"   parsed: {output}"
                    )
                    raise InvalidMappingError(msg)

                return first_result["output"]

        return None

    def map_identities(
        self, identity_data: t.Iterable[t.Mapping[str, str]]
    ) -> list[dict[str, list[str]]]:
        """
        Map all identities from the associated input list to local usernames
        that can potentially access the configured Storage Gateway

        :param identity_data: List of identity assertions from Globus Auth

        :returns:
            List of mappings from the `id` property of the input data
            to a list of local username strings that the identity maps to.

        :raises errors.InternalServerError:
            Error executing the external mapping program, or the output
            from the program is not in an understood format.
        """
        results: list[dict[str, list[str]]] = []

        cmd = self.command + ["-c", self.connector, "-s", self.storage_gateway, "-a"]
        cmd_str = shlex.join(cmd)
        p = Popen(cmd, universal_newlines=True, stdout=PIPE, stderr=PIPE, stdin=PIPE)

        input_data = identity_mapping_input_schema.dumps({"identities": identity_data})
        stdout, stderr = p.communicate(input_data, timeout=self.mapping_timeout_s)
        self.log.debug(
            # fmt: off
            "\n  command: %s"
            "\n    stdin: %s"
            "\n   stdout: %s"
            "\n   stderr: %s",
            # fmt: on
            cmd_str,
            input_data,
            stdout,
            stderr,
        )

        if p.returncode != 0:
            msg = (
                "Error mapping identities\n"
                f"    command: {cmd_str}\n"
                f"  exit code: {p.returncode}\n"
                f"      stdin: {input_data}\n"
                f"     stdout: {stdout}\n"
                f"     stderr: {stderr}"
            )
            raise IdentityMappingError(msg)

        if stdout:
            input_ids = tuple(i["id"] for i in identity_data)
            try:
                output = identity_mapping_output_schema.loads(stdout)
            except Exception as e:
                msg = (
                    "The mapping application yielded an invalid result:\n"
                    f"  exception: ({type(e).__name__}) {e}\n"
                    f"    command: {cmd_str}\n"
                    f"      stdin: {input_data}\n"
                    f"     stdout: {stdout}\n"
                    f"     stderr: {stderr}"
                )
                raise IdentityMappingError(msg) from e

            result: dict[str, list[str]] = {}

            for o in output["result"]:
                output_id = str(o["id"])
                if output_id not in input_ids:
                    raise InvalidMappingError("Invalid ID in result")
                if output_id not in result:
                    result[output_id] = []
                result[output_id].append(o["output"])
            results.append(result)
        return results

    @property
    def log(self) -> logging.Logger:
        if not self._log:
            self._log = logging.getLogger(__file__)
        return self._log

    @log.setter
    def log(self, logger: logging.Logger):
        self._log = logger

    @log.deleter
    def log(self):
        self._log = None



import marshmallow as ma

from .data_type import Schema, data_type, data_type_schema


class MappingExpression(ma.Schema):
    """
    The MappingExpression document type contains information about a mapping
    expression, including the input, match, output, and flags used to process
    this expression.
    """

    source = ma.fields.String(
        metadata={
            "description": (
                "A string comprised of text plus identity set data field names"
                " surrounded by curly brackets `{}` which are interpolated into the"
                " text."
            )
        }
    )
    match = ma.fields.String(
        metadata={
            "description": (
                "An expression which is applied to the output performing"
                " interpolation on source for determining if this mapping applies."
                " This requires a full string match on the source."
            )
        }
    )
    ignore_case = ma.fields.Boolean(
        metadata={
            "description": (
                "Flag indicating the match should be executed as a case insensitive"
                " comparison. If not present, this defaults to false."
            )
        }
    )
    literal = ma.fields.Boolean(
        metadata={
            "description": (
                "Flag indicating the match expression should be done as a literal"
                " match, ignoring any special regular characters. If not present,"
                " this defaults to false."
            )
        }
    )
    output = ma.fields.String(
        metadata={
            "description": (
                "A string representing the result of the mapping if the match"
                " succeeded. References to the original identity_set data can be"
                " interpolated as in the *source* property.  References to match"
                " groups from the *match* property can be interpolated with numbers"
                " (indices starting with 0) surrounded by curly brackets `{}`."
            )
        }
    )


@data_type("expression_identity_mapping#1.0.0", schema_base_name="IdentityMapping")
class ExpressionIdentityMapping_1_0_0(Schema):
    __doc__ = """
    The ExpressionIdentityMapping defines a set of identity mapping expressions
    to map Globus Auth identity data to a connector-specific list of account
    names.
    """

    mappings = ma.fields.List(
        ma.fields.Nested(MappingExpression),
        metadata={"description": "Array of expression-based identity mapping values"},
    )


# We use IdentityMapping in the storage gateway schema like an ordinary
# DATA_TYPE schema, but this one is a bit different since it has disjoint
# data types, in this case, we provide more insight in the docstring and
# choose the desired DATA_TYPE from the object we're serializing
IdentityMapping = data_type_schema.IdentityMapping
IdentityMapping.__doc__ = """
    Globus Connect Server provides two ways for you to implement a custom
    Globus identity to account mapping: expression-based and external program

    With expression-based mapping you can write rules that extract data from
    fields in the Globus identity document to form storage gateway-specific
    usernames. If there is a regular relationship between most of your users'
    Identity information to their account names, this is probably the most
    direct way to accomplish the mapping.

    With external program mappings you can use any mechanism you like (static
    mapping, ldap, database, etc) to look up account information and return the
    mapped account user name. If you have an account system that has usernames
    without a simple relationship to your users' Globus identities, or that
    requires interfacing with an accounting system, this may be necessary.
"""


def __get_identity_mapping_object_type(obj):
    return ma.utils.get_value(obj, "DATA_TYPE")


IdentityMapping.get_obj_type = __get_identity_mapping_object_type



"""
This module implements a mapping algorithm to transform an Identity resource
from Globus Auth to a username which is meaningful in the context of the
Connector that is associated with the Storage Gateway where mapping is being
performed.

The input mapping_data consists of a list of mappings, which are passed in
as a dicts containing the following fields:

*   source: A source expression that contains a string which may include
    interpolations from properties of the Identity resource being
    evaluated.
*   match: A match expression to compare against the source. This
    expression supports a limited regular expression syntax which is
    applied to the source string.
*   output: An output expression which can contain references to properties
    of the input Identity resource or references to parenthesized
    subexpressions from applying the match expression to the source.
*   ignore_case: A boolean flag controlling whether to perform a case
    insensitive match. The default is False.
*   literal: A boolean flag whether to interpet the match expression
    literally, or to use the regular expression language defined below when
    determining whether a match succeeds. If this is set to True,
    references to match subexpressions may not be used in the output field.

Match Expression Language
-------------------------

When the literal flag is not present, the match value is interpreted using
a reduced version of the POSIX extended regular expression, including
special processing for the following tokens:

* :code:`.` any single character match
* :code:`?` for zero or one of the preceding match
* :code:`*` for zero or more of the preceding match
* :code:`|` for alternative match branches
* :code:`()` to group a match and created a numbered submatch for output
  interpolation
* `\\` to escape any of the above, `\\`, or :code:`"` from special
  processing

Any other string in the match expression is matched literally

Identity Data Interpolation
---------------------------

The source and output strings may include values that are interpolated from
a Globus auth identity resource, as described in the
`Globus Auth documentation
<https://docs.globus.org/api/auth/reference/#_identity_resource>`_.

Interpolations are indicated by surrounding the resource property with curly
braces. For example, to interpolate the username property of the identity
resource in the source string, one uses the value :code:`{username}`; likewise,
to interpolate the id property, one uses the string :code:`{id}`. Multiple
values may be interpolated into either the source or output strings.

Subexpression Interpolation
---------------------------

The output string may also contain references to subexpression matches which
are also interpolated into the output. The matching is down by using a similar
syntax the Identity Data Interpolation above, but using integer values to
reference matched subexpression, with indices starting at 0. For example, using
a match expression :code:`(.*)@example\\.org`, one can interpolate the part
before the :code:`@` symbol my using the string :code:`{0}` in the output
expression.

Examples
--------

To match username against a given domain, and return the username as the
output expression:

.. code-block:: json

    {
        "source": "{username}",
        "match": "(.*)@example\\.org",
        "output": "{0}"
    }

To match an identity ID and map to a particular user account

.. code-block:: json

    {
        "source": "{id}",
        "match": "b34bb069-3b68-4d38-b4e6-bc1e4fc95332",
        "output": "application_account"
    }

To rewrite an identity from one domain to another:

.. code-block: json

    {
        "source": "{username}",
        "match": "(.*)@example\\.edu"
        "output": "{0}@data.example.edu"
    }

To map a set of identities to a common account:

.. code-block: json

    {
        "source": "{username}",
        "match": "(alice|bob)@example\\.org",
        "output": "globus"
    }
"""

from __future__ import annotations

import re
import typing as t
from typing_extensions import Self

from .abstract import IdentityMapping
from ..errors import IdentityMappingError, InvalidMappingError
from ..schema import IdentityMappingConfiguration


class ExpressionIdentityMappingDict(t.TypedDict, total=False):
    source: str
    match: str
    output: str
    literal: bool
    ignore_case: bool


class CompiledExpressionIdentityMappingDict(t.TypedDict, total=False):
    source: str
    match: str
    output: str
    literal: bool
    ignore_case: bool
    compiled_match: t.Pattern


class ExpressionMappingDocument(IdentityMappingConfiguration):
    mappings: list[ExpressionIdentityMappingDict]


class ExpressionIdentityMapping(IdentityMapping):
    """
    The ExpressionIdentityMapping class implements a mapping algorithm to
    transform an Identity resource from Globus Auth to a username which is
    meaningful in the context of the Connector that is associated with the
    Storage Gateway mapping is being performed on.
    """

    DATA_TYPE = "expression_identity_mapping#1.0.0"

    def __init__(
        self,
        mapping_data: t.Iterable[ExpressionIdentityMappingDict],
        storage_gateway: str | None = None,
        connector: str | None = None,
    ) -> None:
        """
        :param list[ExpressionIdentityMappingDict] mapping_data:
            Identity mapping definitions as described in the API documentation.
        :param str storage_gateway:
            The storage gateway that the mapping will be performed for. This is
            not used in this module
        :param str connector:
            The connector uuid that the mapping will be performed for. This is
            not used in this module
        """
        self.mappings: list[CompiledExpressionIdentityMappingDict] = []
        self.connector = connector
        self.storage_gateway = storage_gateway

        for m in mapping_data:
            mapping: CompiledExpressionIdentityMappingDict = {
                "source": m["source"],
                "compiled_match": self._compile_match(
                    m["match"],
                    m.get("ignore_case", False),
                    m.get("literal", False),
                ),
                "match": m["match"],
                "output": m["output"],
            }
            if "ignore_case" in m:
                mapping["ignore_case"] = m["ignore_case"]
            if "literal" in m:
                mapping["literal"] = m["literal"]
            self.mappings.append(mapping)
        self._init_mappings = mapping_data

    def __repr__(self) -> str:
        args = ", ".join(map(str, self._init_mappings))
        sg = f"'{self.storage_gateway}'" if self.storage_gateway else "None"
        conn = f"'{self.connector}'" if self.connector else "None"
        return f"{type(self).__name__}([{args}], {sg}, {conn})"

    @classmethod
    def _from_mapping_document(
        cls,
        mapping_document: ExpressionMappingDocument,
        connector: str | None = None,
        storage_gateway: str | None = None,
    ) -> Self:
        return cls(
            mapping_document["mappings"],
            connector=connector,
            storage_gateway=storage_gateway,
        )

    @staticmethod
    def _compile_match(
        match: str, ignore_case: bool = False, literal: bool = False
    ) -> t.Pattern:
        original_match = match
        if not literal:
            bad_escape = re.search(r"(\\[^.?*|()\\0-9])", match)
            if bad_escape:
                raise InvalidMappingError(
                    f"{match} - Invalid \\ expression: {bad_escape.group()}"
                )
            unsupported_re = re.search(r"([*?]\?)", match)
            if unsupported_re:
                raise InvalidMappingError(
                    f"{match} - Invalid ? expression: {unsupported_re.group()}"
                )
            leading_wildcard_re = re.search(r"^([*?])", match)
            if leading_wildcard_re:
                _g = leading_wildcard_re.group()
                raise InvalidMappingError(
                    f"{match} - Invalid leading wildcard expression: {_g}"
                )

            # Escape python regex specials that we don't support
            match = re.sub(r"(?<!\\)([\^\$\+\{\}\[\]])", r"\\\1", match)
        else:
            match = re.sub(r"(?<!\\)([\^\?\*\|\(\)\\\$\+\.\{\}\[\]])", r"\\\1", match)
            match = re.sub(r"(\\[0-9])", r"\\\1", match)

        try:
            return re.compile("^" + match + "$", re.I if ignore_case else 0)
        except Exception as e:
            raise InvalidMappingError("Invalid mapping {}: {}", original_match, str(e))

    @staticmethod
    def _interpolate(
        string: str, identity_data: t.Mapping[str, str], matches: list | None = None
    ) -> str:
        if not matches:
            matches = []
        interpolated = string

        for m in re.finditer(r"(?<!\\)(\\\\)*({([^}]*)})", string):
            subexp, subname = m.groups()[1:]
            try:
                matchref = int(subname)
                if matchref >= len(matches):
                    raise IdentityMappingError(f"Invalid reference to {matchref}")
                repl = matches[matchref]
            except ValueError:
                repl = identity_data.get(subname, None)
            if repl is None:
                raise IdentityMappingError(f"Missing identity property {subname}")
            interpolated = interpolated.replace(subexp, repl)
        return interpolated

    def map_identity(self, identity_data: t.Mapping[str, str]) -> str | None:
        """
        Applies the defined mapping (e.g., as defined via
        `.from_mapping_document()`) to `identity_data`, and, if successful,
        returns the interpolated output string of the first match.  If no
        mapping matches, returns `None`.
        """
        for m in self.mappings:
            source = self._interpolate(m["source"], identity_data)
            mr = m["compiled_match"].match(source)
            if mr:
                return self._interpolate(m["output"], identity_data, list(mr.groups()))
        return None

    def map_identities(
        self, identity_data: t.Iterable[t.Mapping[str, str]]
    ) -> list[dict[str, list[str]]]:
        """
        Applies each defined mapping -- in order -- to each identity resource
        in `identity_data`.  Matches are included in the result.  Each item in
        the returned list is a dictionary mapping the id of an identity to a
        list of successfully mapped values.

        In pseudocode:

            for mapping in mappings:
                for identity in identity_data:
                    if mapping matches identity:
                        add to output
        """
        results = []
        for m in self.mappings:
            m_re = m["compiled_match"]
            m_output = m["output"]
            mapping_results: dict[str, list[str]] = {}
            for rec in identity_data:
                if rec.get("status") not in ("used", "private"):
                    continue

                source = self._interpolate(m["source"], rec)
                mr = m_re.match(source)
                if mr:
                    identity_id = rec["id"]
                    if identity_id not in mapping_results:
                        mapping_results[identity_id] = []
                    mapping_results[identity_id].append(
                        self._interpolate(m_output, rec, list(mr.groups()))
                    )
            results.append(mapping_results)
        return results



class InvalidMappingError(ValueError):
    """
    General errors related to the acquisition of an identity mapping; for
    example, if an external program returns an incorrectly-formatted
    response or if the response does not match the input request.
    """


class IdentityMappingError(ValueError):
    """
    General errors related to invoking the identity mapping logic; for
    example, if an external program returns an error.
    """



from __future__ import annotations

import importlib.metadata
import logging
import sys
import typing as t

from .protocol import IdentityMappingProtocol
from .errors import InvalidMappingError

log = logging.getLogger(__name__)


def load_mappers(
    mappings_list: t.Iterable[dict | IdentityMappingProtocol] | None,
    connector: str | None = None,
    storage_gateway: str | None = None,
) -> list[IdentityMappingProtocol]:
    """
    Dynamically instantiate IdentityMapping objects based on the passed
    configuration documents.

    An IdentityMapping configuration document is just a dictionary with a
    structure as outlined in Globus Connect Server's Identity Mapping Guide
    documentation[1].  That's basically a structure:

        {
            "DATA_TYPE": "<class_name>#<version>"
            "further": [
                "structure per the specific class' specification"
                " and needs"
            ]
        }

    An example use of this helper function might look like:

        def get_mappers_from_application_specific_config(...):
            config_doc_path = pathlib.Path("path/to/identity_config.json")
            config_document_texts = config_doc_path.read_text()
            config_documents = json.loads(config_document_texts)
            return load_mappers(config_documents, connector_id, storage_gateway_id)

    A list of Globus-specific Connector Identifiers is also available in the
    Identity Mapping Guide[1].  (See "Command Line Options")

    After loading, the mappers may each be queried to determine a context-
    appropriate username.  That might look like:

        for mapper in mappers:
            username = mapper.map_identity(globus_identity_object)

    [1] https://docs.globus.org/globus-connect-server/v5.4/identity-mapping-guide/
    """
    mappings: list[IdentityMappingProtocol] = []
    for ndx, map_or_config in enumerate(mappings_list or []):
        if isinstance(map_or_config, dict):
            data_type: str | None = map_or_config.get("DATA_TYPE")
            if not data_type:
                raise InvalidMappingError(f"Invalid map config (index: {ndx})")

            mapping_identifier = data_type.split("#", 1)[0]
            mapper = None
            eps = importlib.metadata.entry_points()
            if sys.version_info < (3, 10):
                gim_lib = eps.get("globus_identity_mapping", [])
            else:
                gim_lib = eps.select(group="globus_identity_mapping")
            for ep in gim_lib:
                if ep.name == mapping_identifier:
                    entry_point = ep.load()
                    mapper = entry_point.from_mapping_document(
                        map_or_config,
                        connector=connector,
                        storage_gateway=storage_gateway,
                    )
                    break
            if mapper:
                mappings.append(mapper)
                continue
            log.warning(f"No identity mapper found for: {data_type} (index: {ndx})")

        elif hasattr(map_or_config, "map_identity"):
            mappings.append(map_or_config)
        else:
            raise InvalidMappingError(f"Invalid mapping or map config (index: {ndx})")

    return mappings


__all__ = ("load_mappers",)
